/// <reference types="node" />
/// <reference types="node" />
import { EventEmitter } from "events";
/**
 * Key-validation: The keys can be given as either string or number,
 * but are casted to a string internally anyway.
 */
export type Key = string | number;
type Data<VT> = {
    [keyof: Key]: WrappedValue<VT>;
};
type WrappedValue<VT> = {
    t: number;
    v: VT | string;
};
export type Options = {
    /**
     * If enabled, all values will be stringified during the set operation
     *
     * @type {boolean}
     * @memberof Options
     */
    forceString?: boolean;
    /**
     * used standard size for calculating value size of object
     */
    objectValueSize?: number;
    /**
     * used standard size for calculating value size of promise
     */
    promiseValueSize?: number;
    /**
     * used standard size for calculating value size of array
     */
    arrayValueSize?: number;
    /**
     * standard time to live in seconds. 0 = infinity
     *
     * @type {number}
     * @memberof Options
     */
    stdTTL?: number;
    /**
     * time in seconds to check all data and delete expired keys
     *
     * @type {number}
     * @memberof Options
     */
    checkperiod?: number;
    /**
     * en/disable cloning of variables.
     * disabling this is strongly encouraged when aiming for performance!
     *
     * If `true`: set operations store a clone of the value and get operations will create a fresh clone of the cached value
     * If `false` you'll just store a reference to your value
     *
     * @type {boolean}
     * @memberof Options
     */
    useClones?: boolean;
    /**
     * whether values should be deleted automatically at expiration
     *
     * @type {boolean}
     */
    deleteOnExpire?: boolean;
    /**
     * max amount of keys that are being stored.
     * set operations will throw an error when the cache is full
     *
     * @type {number}
     * @memberof Options
     */
    maxKeys?: number;
};
export default class NodeCache<VT> extends EventEmitter {
    options: Required<Options>;
    data: Data<VT>;
    stats: {
        hits: number;
        misses: number;
        keys: number;
        ksize: number;
        vsize: number;
    };
    validKeyTypes: string[];
    checkTimeout: NodeJS.Timeout | null;
    constructor(options?: Options);
    get(key: Key): VT | string | undefined;
    mget(keys: Key[]): {
        [keyof: string]: string | VT;
        [keyof: number]: string | VT;
    };
    set(key: Key, value: VT, ttl?: number): boolean;
    fetch(key: Key, value: VT | (() => VT), ttl: number | string | undefined): string | VT | undefined;
    mset(keyValueSet: {
        key: number | string;
        val: VT;
        ttl: number;
    }[]): boolean;
    del(keys: Key | Key[]): number;
    take(key: Key): string | VT | undefined;
    ttl(key: Key, ttl: number): boolean;
    getTtl(key: Key): number | undefined;
    keys(): string[];
    has(key: Key): boolean;
    getStats(): {
        hits: number;
        misses: number;
        keys: number;
        ksize: number;
        vsize: number;
    };
    flushAll(_startPeriod?: boolean): void;
    flushStats(): void;
    close(): void;
    _checkData(startPeriod?: boolean): Promise<void>;
    _killCheckPeriod(): void;
    _check(key: Key, data: WrappedValue<VT> | undefined): boolean;
    _isInvalidKey(key: unknown): Error | undefined;
    _wrap(value: VT | string, ttl: number | string, asClone?: boolean): WrappedValue<VT>;
    _unwrap(value: WrappedValue<VT>, asClone?: boolean): VT | string;
    _getKeyLength(key: number | string): number;
    _getValLength(value: VT | string): number;
    _error(name: string, message?: string): Error;
}
export {};
